/*
dots "voxel" landscape viewer/points generator targeted at Acorn Archimedes ARM2 8mhz hardware

this is not raycasting but simple dots positioned as a grid with rough perspective layout, dots are then expanded as blocks (16xN) and shaded according to a colormap, dots Y position is modulated by a heightmap

it use fixed point arithmetic because i thought of computing the points at first but for performances reason i switched to precomputed points (computing points has zero advantages except less data)

total visible points and total visible points divided by 3 is shown in top left corner once data is generated, theses are used as parameters for the ARM2 landscape code, total points should be divisible by 3 for it to work on the Archi

can be adjusted for more powerful hardware by increasing the resolution (may need to adjust the points grid)

heightmap and colormap needs to be of the same size

LEFT CLICK = generate points data, total points should be divisible by 3 for Archimedes
RIGHT CLICK = download points data
*/

let heightmap = null
let colormap = null
function preload() {
  // 128x128 heightmap / colormap (maps from Comanche Maximum Ovekill, 90s games)
  heightmap = loadImage('D1.png');
  colormap = loadImage('C1W.png');
}

function setup() {
  createCanvas(320, 256);
  
  heightmap.loadPixels();
  colormap.loadPixels();
}

let vt = []
let points = 0
let binData = []

function draw() {
  noFill()
  noStroke()

  // sky gradient / clearing the top part
  for (let y = 0; y < height / 2.25; y += 1) {
    let b = 255 * (y / (height / 4))
    fill(b,b,b)
    rect(0, y, width, 1)
  }
    
  // = landscape viewer/generator
  let mx = 0 // camera X position
  let my = 0 - frameCount / 4 // camera Y position (+ scroll)
  
  // landscape resolution
  let stepx = 8
  let stepy = 8
  let gridSpacing = 32
  let heightFactor = 0.4
  
  for (let y = 0; y < height; y += stepy) {
    for (let x = 0; x < width; x += stepx) {
      // generate the grid of points (fixed point arithmetic)
      let xx = x << 16
      let yy = y << 16
      
      // go into -1,1 for x
      xx = ((width / 2) << 16) - xx
      
      // fake perspective grid layout
      xx = xx + floor(xx / gridSpacing) * floor(y / stepy)
      
      xx >>= 16
      yy >>= 16
      
      let z = 2-min(1, y / height * (yy/3)) // will be used to dim the farthest points (only affect the viewer, not implemented on Archimedes)
      
      // colormap/heightmap lookup (this also take care of wraparound)
      let heightmapIndex = (floor(x / stepx + mx) % heightmap.width + (floor(heightmap.height + ((y / stepy + my) % heightmap.height)) % heightmap.height) * heightmap.width) * 4
      let hc = heightmap.pixels[heightmapIndex + 0]
      
      // get point colors
      let r = colormap.pixels[heightmapIndex + 0] * z
      let g = colormap.pixels[heightmapIndex + 1] * z
      let b = colormap.pixels[heightmapIndex + 2] * z
      let h = hc * heightFactor // compute height
    
      // center the grid and elevate points (+ do more perspective tweaking)
      xx = width / 2 + xx
      let perspTweaks = yy / 4 - h + (y / height) * (yy / 2)
      yy = height / 2 + perspTweaks
      yy -= 16 // adjust vertical position of the landscape
      
      if (xx > -12 && xx < width) { // filter points (this is used to tweak final points count so that total number of points are divisible by 3, some points should be left overflowing to avoid gaps on the sides, they will be clamped)
        xx = Math.min(xx, width-4)
        
        // for clipping we make the overflowing points fixed to a 4px grid to simplify it
        if (xx >= width-16) xx = floor(xx / 4) * 4
        if (xx< 0) xx = floor(xx / 4) * 4
        
        // this part is the points generator (for Archimedes version)
        // this part accumulate all visible points in a visibility table, this filter out points that will never be visible
        let v = vt[x + y * width]
        if (!v) {
          if (yy >= height-1) { // point will never be visible so don't bother
           vt[x + y * width] = 0
          } else {
            vt[x + y * width] = {
               // compute the screen position, we just want the flat plane (not the elevated one) so also undo the elevation
              xy: Math.max(xx, 0) + round(yy + h) * width,
              // compute U/V for the point (as a maps data address)
              uv: (floor(x/stepx) % heightmap.width + (floor(heightmap.height + ((y/stepy)%heightmap.height)) % heightmap.height) * heightmap.width)*8, // 8 because map data is 2 words (height,color), see the associated sketch
              x: xx
            }
          }
        }

        fill(r, g, b)
        rect(xx, yy,16,11)
      }
    }
  }

  fill(255)
  textSize(16)
  textAlign(LEFT)
  text("total points: " + points, 16, 16)
  text("   points / 3: " + points/3, 16, 32)
  
  textAlign(CENTER)
  text("Right click to download points data", width / 2, height - 2)
}

// generate / download Archimedes points data on click
function mouseReleased() {
  // generate points data (interleaved x/y/u/v + 4-bit clipping word)
  if (mouseButton === LEFT) {
    binData = []
    points = 0

    let xStack = []
    for (let i = 0; i < vt.length; i += 1) {
      if (vt[i]) {
        points += 1
        binData.push(vt[i].xy)
        binData.push(vt[i].uv)
        xStack.push(vt[i].x)

        if (points % 3 === 0) {
          // clipping left & right (4-bit value which indicate which block width to use)
          let nBlocks1 = 3
          let nBlocks2 = 3
          let nBlocks3 = 3
          if (xStack[0] < 0) {
            nBlocks1 = floor(xStack[0] + 16) / 4
          }
          if (xStack[1] < 0) {
            nBlocks2 = floor(xStack[1] + 16) / 4
          }
          if (xStack[2] < 0) {
            nBlocks3 = floor(xStack[2] + 16) / 4
          }
          if (xStack[0] >= width-16) {
            nBlocks1 = 3-floor(xStack[0] + 16) % width / 4
          }
          if (xStack[1] >= width-16) {
            nBlocks2 = 3-floor(xStack[1] + 16) % width / 4
          }
          if (xStack[2] >= width-16) {
            nBlocks3 = 3-floor(xStack[2] + 16) % width / 4
          }

          let nBlocks = nBlocks1 | (nBlocks2 << 2) | (nBlocks3 << 4)
          binData.push(nBlocks)

          xStack = []
        }
      }
    }
  }

  if (mouseButton === RIGHT) {
    // download stuff
    const buffer = new ArrayBuffer(binData.length * 4);
    const vue32 = new Int32Array(buffer);
    vue32.set(binData);

    const link = document.createElement( 'a' );
    link.style.display = 'none';
    document.body.appendChild( link );

    const blob = new Blob( [ buffer ], { type: "application/octet-stream" } );	
    const objectURL = URL.createObjectURL( blob );

    link.href = objectURL;
    link.href = URL.createObjectURL( blob );
    link.download = 'points,ffd';

    link.click();  
  }
}


